import importlib
import inspect
import logging
import sys
import traceback
from enum import Enum
from pathlib import Path
from threading import Lock, Thread
from typing import Dict, Union

import kserve
import numpy as np
from cloudevents.http import CloudEvent
from common.util import assign_request_to_inputs_instances_after_validation
from kserve.errors import InvalidInput
from kserve.grpc.grpc_predict_v2_pb2 import ModelInferRequest, ModelInferResponse
from shared.secrets_resolver import SecretsResolver

MODEL_BASENAME = "model"


class ModelWrapper(kserve.Model):
    class Status(Enum):
        NOT_READY = 0
        LOADING = 1
        READY = 2
        FAILED = 3

    _config: dict
    _model: object
    _load_lock: Lock = Lock()
    _predict_lock: Lock = Lock()
    _status: Status = Status.NOT_READY
    _logger: logging.Logger

    def __init__(self, config: dict):
        super().__init__(MODEL_BASENAME)
        self._config = config
        self.logger = logging.getLogger(__name__)

    def load(self) -> bool:
        if self.ready:
            return self.ready

        # if we are already loading, just pass; our container will return 503 while we're loading
        if not self._load_lock.acquire(blocking=False):
            return False

        self._status = ModelWrapper.Status.LOADING

        self.logger.info("Executing model.load()...")

        try:
            self.try_load()
            self.ready = True
            self._status = ModelWrapper.Status.READY

            self.logger.info("Completed model.load() execution")

            return self.ready
        except Exception:
            self.logger.exception("Exception while loading model")
            self._status = ModelWrapper.Status.FAILED
        finally:
            self._load_lock.release()

        return self.ready

    def start_load(self):
        if self.should_load():
            thread = Thread(target=self.load)
            thread.start()

    def load_failed(self) -> bool:
        return self._status == ModelWrapper.Status.FAILED

    def should_load(self) -> bool:
        # don't retry failed loads
        return (
            not self._load_lock.locked()
            and not self._status == ModelWrapper.Status.FAILED
            and not self.ready
        )

    def try_load(self):
        if "bundled_packages_dir" in self._config:
            bundled_packages_path = Path("/packages")
            if bundled_packages_path.exists():
                sys.path.append(str(bundled_packages_path))
        model_module_name = str(
            Path(self._config["model_class_filename"]).with_suffix("")
        )
        module = importlib.import_module(
            f"{self._config['model_module_dir']}.{model_module_name}"
        )
        model_class = getattr(module, self._config["model_class_name"])
        model_class_signature = inspect.signature(model_class)
        model_init_params = {}
        if _signature_accepts_keyword_arg(model_class_signature, "config"):
            model_init_params["config"] = self._config
        if _signature_accepts_keyword_arg(model_class_signature, "data_dir"):
            model_init_params["data_dir"] = Path("data")
        if _signature_accepts_keyword_arg(model_class_signature, "secrets"):
            model_init_params["secrets"] = SecretsResolver.get_secrets(self._config)
        self._model = model_class(**model_init_params)

        if hasattr(self._model, "load"):
            self._model.load()

    def validate(self, payload):
        if (
            "instances" in payload
            and not isinstance(payload["instances"], (list, np.ndarray))
            or "inputs" in payload
            and not isinstance(payload["inputs"], (list, np.ndarray))
        ):
            raise InvalidInput(
                'Expected "instances" or "inputs" to be a list or NumPy ndarray'
            )

        return assign_request_to_inputs_instances_after_validation(payload)

    def preprocess(
        self,
        payload: Union[Dict, CloudEvent, ModelInferRequest],
        headers: Dict[str, str] = None,
    ) -> Union[Dict, ModelInferRequest]:
        if not hasattr(self._model, "preprocess"):
            return payload
        return self._model.preprocess(payload)

    def postprocess(
        self, response: Union[Dict, ModelInferResponse], headers: Dict[str, str] = None
    ) -> Dict:
        if not hasattr(self._model, "postprocess"):
            return response
        return self._model.postprocess(response)

    def predict(
        self, payload: Union[Dict, ModelInferRequest], headers: Dict[str, str] = None
    ) -> Union[Dict, ModelInferResponse]:
        try:
            self._predict_lock.acquire()
            return self._model.predict(payload)
        except Exception:
            response = {}
            logging.exception("Exception while running predict")
            response["error"] = {"traceback": traceback.format_exc()}
            return response
        finally:
            self._predict_lock.release()


def _signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False



def model_supports_predict_proba(model: object) -> bool:
    if not hasattr(model, "predict_proba"):
        return False
    if hasattr(
        model, "_check_proba"
    ):  # noqa eg Support Vector Machines *can* predict proba if they made certain choices while training
        try:
            model._check_proba()
            return True
        except AttributeError:
            return False
    return True


def assign_request_to_inputs_instances_after_validation(body: dict) -> dict:
    # we will treat "instances" and "inputs" the same
    if "instances" in body and "inputs" not in body:
        body["inputs"] = body["instances"]
    elif "inputs" in body and "instances" not in body:
        body["instances"] = body["inputs"]
    return body



from pathlib import Path

from truss import __version__

# This needs to be updated whenever new truss base images are published. This
# manual nature is by design; at this point we want any new base image releases
# to be reviewed via code change before landing. This value should be typically
# set to the latest version of truss library.
#
# [IMPORTANT] Make sure all images for this version are published to dockerhub
# before change to this value lands. This value is used to look for base images
# when building docker image for a truss.
TRUSS_BASE_IMAGE_VERSION_TAG = "v0.3.1"


def file_is_empty(path: Path, ignore_hash_style_comments: bool = True) -> bool:
    if not path.exists():
        return True

    with path.open() as file:
        for line in file.readlines():
            if ignore_hash_style_comments and _is_hash_style_comment(line):
                continue
            if line.strip() != "":
                return False

    return True


def file_is_not_empty(path: Path, ignore_hash_style_comments: bool = True) -> bool:
    return not file_is_empty(path, ignore_hash_style_comments)


def _is_hash_style_comment(line: str) -> bool:
    return line.lstrip().startswith("#")


def truss_base_image_name(job_type: str) -> str:
    return f"baseten/truss-{job_type}-base"


def truss_base_image_tag(
    python_version: str,
    use_gpu: bool,
    live_reload: bool,
    version_tag: str = None,
) -> str:
    if version_tag is None:
        version_tag = f"v{__version__}"

    base_tag = python_version
    if use_gpu:
        base_tag = f"{base_tag}-gpu"
    if live_reload:
        base_tag = f"{base_tag}-reload"
    return f"{base_tag}-{version_tag}"


def to_dotted_python_version(truss_python_version: str) -> str:
    """Converts python version string using in truss config to the conventional dotted form.

    e.g. py39 to 3.9
    """
    return f"{truss_python_version[2]}.{truss_python_version[3:]}"



import pickle
from typing import Dict, List


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        self._config = kwargs["config"]
        self._pipeline = None

    def load(self):
        with open(self._data_dir / "pipeline.cpick", "rb") as f:
            self._pipeline = pickle.load(f)

    def preprocess(self, request: Dict) -> Dict:
        """
        Incorporate pre-processing required by the model if desired here.

        These might be feature transformations that are tightly coupled to the model.
        """
        return request

    def postprocess(self, request: Dict) -> Dict:
        """
        Incorporate post-processing required by the model if desired here.
        """
        return request

    def predict(self, request: Dict) -> Dict[str, List]:
        return self._pipeline(request)



from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader
from truss.constants import (
    CONTROL_SERVER_CODE_DIR,
    MODEL_DOCKERFILE_NAME,
    MODEL_README_NAME,
    REQUIREMENTS_TXT_FILENAME,
    SERVER_CODE_DIR,
    SERVER_DOCKERFILE_TEMPLATE_NAME,
    SERVER_REQUIREMENTS_TXT_FILENAME,
    SHARED_SERVING_AND_TRAINING_CODE_DIR,
    SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
    SYSTEM_PACKAGES_TXT_FILENAME,
    TEMPLATES_DIR,
)
from truss.contexts.image_builder.image_builder import ImageBuilder
from truss.contexts.image_builder.util import (
    TRUSS_BASE_IMAGE_VERSION_TAG,
    file_is_not_empty,
    to_dotted_python_version,
    truss_base_image_name,
    truss_base_image_tag,
)
from truss.contexts.truss_context import TrussContext
from truss.patch.hash import directory_content_hash
from truss.readme_generator import generate_readme
from truss.truss_spec import TrussSpec
from truss.utils import build_truss_target_directory, copy_file_path, copy_tree_path

BUILD_SERVER_DIR_NAME = "server"
BUILD_CONTROL_SERVER_DIR_NAME = "control"


class ServingImageBuilderContext(TrussContext):
    @staticmethod
    def run(truss_dir: Path):
        return ServingImageBuilder(truss_dir)


class ServingImageBuilder(ImageBuilder):
    def __init__(self, truss_dir: Path) -> None:
        self._truss_dir = truss_dir
        self._spec = TrussSpec(truss_dir)

    @property
    def default_tag(self):
        return f"{self._spec.model_framework_name}-model:latest"

    def prepare_image_build_dir(self, build_dir: Path = None):
        """Prepare a directory for building the docker image from.

        Returns:
            docker command to build the docker image.
        """
        if build_dir is None:
            build_dir = build_truss_target_directory(self._spec.model_framework_name)
            # todo: Add a logging statement here, suggesting how to clean up the directory.

        copy_tree_path(self._spec.truss_dir, build_dir)
        copy_tree_path(
            SERVER_CODE_DIR,
            build_dir / BUILD_SERVER_DIR_NAME,
        )
        copy_tree_path(
            SHARED_SERVING_AND_TRAINING_CODE_DIR,
            build_dir
            / BUILD_SERVER_DIR_NAME
            / SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
        )
        if self._spec.config.live_reload:
            copy_tree_path(
                CONTROL_SERVER_CODE_DIR,
                build_dir / BUILD_CONTROL_SERVER_DIR_NAME,
            )
        server_reqs_filepath = (
            TEMPLATES_DIR / self._spec.model_framework_name / REQUIREMENTS_TXT_FILENAME
        )
        should_install_server_requirements = (
            server_reqs_filepath.exists() and file_is_not_empty(server_reqs_filepath)
        )
        if should_install_server_requirements:
            copy_file_path(
                TEMPLATES_DIR
                / self._spec.model_framework_name
                / REQUIREMENTS_TXT_FILENAME,
                build_dir / SERVER_REQUIREMENTS_TXT_FILENAME,
            )

        with (build_dir / REQUIREMENTS_TXT_FILENAME).open("w") as req_file:
            req_file.write(self._spec.requirements_txt)

        with (build_dir / SYSTEM_PACKAGES_TXT_FILENAME).open("w") as req_file:
            req_file.write(self._spec.system_packages_txt)

        data_dir_exists = (build_dir / self._spec.config.data_dir).exists()
        bundled_packages_dir_exists = (
            build_dir / self._spec.config.bundled_packages_dir
        ).exists()

        template_loader = FileSystemLoader(str(TEMPLATES_DIR))
        template_env = Environment(loader=template_loader)
        dockerfile_template = template_env.get_template(SERVER_DOCKERFILE_TEMPLATE_NAME)
        config = self._spec.config

        base_image_name = truss_base_image_name(job_type="server")
        tag = truss_base_image_tag(
            python_version=to_dotted_python_version(config.python_version),
            use_gpu=config.resources.use_gpu,
            live_reload=config.live_reload,
            version_tag=TRUSS_BASE_IMAGE_VERSION_TAG,
        )
        base_image_name_and_tag = f"{base_image_name}:{tag}"
        should_install_system_requirements = file_is_not_empty(
            build_dir / SYSTEM_PACKAGES_TXT_FILENAME
        )
        should_install_requirements = file_is_not_empty(
            build_dir / REQUIREMENTS_TXT_FILENAME
        )
        dockerfile_contents = dockerfile_template.render(
            should_install_server_requirements=should_install_server_requirements,
            base_image_name_and_tag=base_image_name_and_tag,
            should_install_system_requirements=should_install_system_requirements,
            should_install_requirements=should_install_requirements,
            config=config,
            data_dir_exists=data_dir_exists,
            bundled_packages_dir_exists=bundled_packages_dir_exists,
            truss_hash=directory_content_hash(self._truss_dir),
        )
        docker_file_path = build_dir / MODEL_DOCKERFILE_NAME
        with docker_file_path.open("w") as docker_file:
            docker_file.write(dockerfile_contents)

        readme_file_path = build_dir / MODEL_README_NAME
        try:
            readme_contents = generate_readme(self._spec)
            with readme_file_path.open("w") as readme_file:
                readme_file.write(readme_contents)
        except Exception as e:
            click.echo(
                click.style(
                    f"""WARNING: Auto-readme generation has failed.
                    This is probably due to a malformed config.yaml or
                    malformed examples.yaml. Error is:
                    {e}
                    """,
                    fg="yellow",
                )
            )



from unittest import mock

from truss.templates.server.common.util import (
    assign_request_to_inputs_instances_after_validation,
)


def test_assign_request_to_inputs_instances_after_validation():
    inputs_input = [1, 2, 3, 4]
    inputs_dict = {"inputs": inputs_input}
    instances_input = [5, 6, 7, 8]
    instances_dict = {"instances": instances_input}

    processed_inputs = assign_request_to_inputs_instances_after_validation(inputs_dict)
    processed_instances = assign_request_to_inputs_instances_after_validation(
        instances_dict
    )

    assert processed_inputs["instances"] == processed_inputs["inputs"] == inputs_input
    assert (
        processed_instances["instances"]
        == processed_instances["inputs"]
        == instances_input
    )


def model_supports_predict_proba():
    mock_not_predict_proba = mock.Mock(name="mock_not_predict_proba")
    mock_not_predict_proba.predict_proba.return_value = False

    mock_check_proba = mock.Mock(name="mock_check_proba")
    mock_check_proba.predict_proba.return_value = True
    mock_check_proba._check_proba.return_value = True

    mock_not_check_proba = mock.Mock(name="mock_not_check_proba")
    mock_not_check_proba.predict_proba.return_value = True
    mock_not_check_proba._check_proba.side_effect = AttributeError

    assert not model_supports_predict_proba(mock_not_predict_proba)
    assert model_supports_predict_proba(mock_check_proba)
    assert not model_supports_predict_proba(mock_not_check_proba)



from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import pkg_resources
import yaml
from truss.patch.hash import file_content_hash_str
from truss.patch.types import TrussSignature
from truss.templates.control.control.helpers.types import (
    Action,
    ModelCodePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import TrussConfig
from truss.truss_spec import TrussSpec


def calc_truss_patch(
    truss_dir: Path, previous_truss_signature: TrussSignature
) -> Optional[List[Patch]]:
    """
    Calculate patch for a truss from a previous state.

    Returns: None if patch cannot be calculated, otherwise a list of patches.
        Note that the none return value is pretty important, patch coverage
        is limited and this usually indicates that the identified change cannot
        be expressed with currently supported patches.
    """
    changed_paths = _calc_changed_paths(
        truss_dir, previous_truss_signature.content_hashes_by_path
    )
    # TODO(pankaj) Calculate model code patches only for now, add config changes
    # later.
    truss_spec = TrussSpec(truss_dir)
    model_module_path = str(truss_spec.model_module_dir.relative_to(truss_dir))
    training_module_path = str(truss_spec.training_module_dir.relative_to(truss_dir))

    patches = []
    for path in changed_paths["removed"]:
        if path.startswith(model_module_path):
            relative_to_model_module_path = str(
                Path(path).relative_to(model_module_path)
            )
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=Action.REMOVE,
                        path=relative_to_model_module_path,
                    ),
                )
            )
        elif path.startswith(training_module_path):
            # Ignore training changes from patch
            continue
        else:
            return None

    for path in changed_paths["added"] + changed_paths["updated"]:
        if path.startswith(model_module_path):
            full_path = truss_dir / path
            relative_to_model_module_path = str(
                Path(path).relative_to(model_module_path)
            )

            # TODO(pankaj) Add support for empty directories, skip them for now.
            if not full_path.is_file():
                continue

            with full_path.open() as file:
                content = file.read()
            action = Action.ADD if path in changed_paths["added"] else Action.UPDATE
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=action,
                        path=relative_to_model_module_path,
                        content=content,
                    ),
                )
            )
        elif path.startswith(training_module_path):
            # Ignore training changes from patch
            continue
        elif str(path) == "config.yaml":
            new_config = TrussConfig.from_yaml(truss_dir / "config.yaml")
            prev_config = TrussConfig.from_dict(
                yaml.safe_load(previous_truss_signature.config)
            )
            config_patches = _calc_config_patches(prev_config, new_config)
            if config_patches is None:
                return None
            patches.extend(config_patches)
        else:
            return None
    return patches


def _calc_changed_paths(
    root: Path, previous_root_path_content_hashes: Dict[str, str]
) -> dict:
    """
    TODO(pankaj) add support for directory creation in patch
    """
    root_relative_paths = set(
        (str(path.relative_to(root)) for path in root.glob("**/*"))
    )
    previous_root_relative_paths = set(previous_root_path_content_hashes.keys())

    added_paths = root_relative_paths - previous_root_relative_paths
    removed_paths = previous_root_relative_paths - root_relative_paths

    updated_paths = set()
    common_paths = root_relative_paths.intersection(previous_root_relative_paths)
    for path in common_paths:
        full_path: Path = root / path
        if full_path.is_file():
            content_hash = file_content_hash_str(full_path)
            previous_content_hash = previous_root_path_content_hashes[path]
            if content_hash != previous_content_hash:
                updated_paths.add(path)

    return {
        "added": list(added_paths),
        "updated": list(updated_paths),
        "removed": list(removed_paths),
    }


def _calc_config_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> Optional[List[Patch]]:
    """Calculate patch based on changes to config.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    # Currently only calculate patches for python requirements and system
    # packages, bail out if anything else has changed.
    if not _only_expected_config_differences(prev_config, new_config):
        return None

    python_requirement_patches = _calc_python_requirements_patches(
        prev_config, new_config
    )
    system_package_patches = _calc_system_packages_patches(prev_config, new_config)
    return [*python_requirement_patches, *system_package_patches]


def _calc_python_requirements_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> Optional[List[Patch]]:
    """Calculate patch based on changes to python requirements.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    patches = []
    prev_reqs = _parsed_reqs_by_name(prev_config.requirements)
    prev_req_names = set(prev_reqs.keys())
    new_reqs = _parsed_reqs_by_name(new_config.requirements)
    new_req_names = set(new_reqs.keys())
    removed_reqs = prev_req_names.difference(new_req_names)
    for removed_req in removed_reqs:
        patches.append(_mk_python_requirement_patch(Action.REMOVE, removed_req))

    added_reqs = new_req_names.difference(prev_req_names)
    for added_req in added_reqs:
        patches.append(
            _mk_python_requirement_patch(Action.ADD, str(new_reqs[added_req]))
        )

    for req in new_req_names.intersection(prev_req_names):
        if new_reqs[req] != prev_reqs[req]:
            patches.append(
                _mk_python_requirement_patch(Action.UPDATE, str(new_reqs[req]))
            )

    return patches


def _calc_system_packages_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> Optional[List[Patch]]:
    """Calculate patch based on changes to system packates.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    patches = []
    prev_pkgs = _system_pacakges_set(prev_config)
    new_pkgs = _system_pacakges_set(new_config)
    removed_pkgs = prev_pkgs.difference(new_pkgs)
    for removed_pkg in removed_pkgs:
        patches.append(_mk_system_package_patch(Action.REMOVE, removed_pkg))

    added_pkgs = new_pkgs.difference(prev_pkgs)
    for added_pkg in added_pkgs:
        patches.append(_mk_system_package_patch(Action.ADD, added_pkg))

    return patches


def _system_pacakges_set(config: TrussConfig) -> Set[str]:
    pkgs = []
    for sys_pkg_line in config.system_packages:
        pkgs.extend(sys_pkg_line.strip().split())
    return set(pkgs)


def _parsed_reqs_by_name(reqs: List[str]) -> Dict[str, Any]:
    parsed_reqs_by_name = {}
    for req in reqs:
        parsed_req = pkg_resources.Requirement.parse(req)
        parsed_reqs_by_name[parsed_req.name] = parsed_req
    return parsed_reqs_by_name


def _only_expected_config_differences(
    prev_config: TrussConfig, new_config: TrussConfig
) -> bool:
    prev_config_dict = prev_config.to_dict()
    prev_config_dict["requirements"] = []
    prev_config_dict["system_packages"] = []

    new_config_dict = new_config.to_dict()
    new_config_dict["requirements"] = []
    new_config_dict["system_packages"] = []

    return prev_config_dict == new_config_dict


def _mk_python_requirement_patch(action: Action, requirement: str) -> Patch:
    return Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=action,
            requirement=requirement,
        ),
    )


def _mk_system_package_patch(action: Action, package: str) -> Patch:
    return Patch(
        type=PatchType.SYSTEM_PACKAGE,
        body=SystemPackagePatch(
            action=action,
            package=package,
        ),
    )



import datetime
import decimal
import json
import uuid


# mostly cribbed from django.core.serializer.DjangoJSONEncoder
def truss_msgpack_encoder(obj, chain=None):
    if isinstance(obj, datetime.datetime):
        r = obj.isoformat()
        if r.endswith("+00:00"):
            r = r[:-6] + "Z"
        return {b"__dt_datetime_iso__": True, b"data": r}
    elif isinstance(obj, datetime.date):
        r = obj.isoformat()
        return {b"__dt_date_iso__": True, b"data": r}
    elif isinstance(obj, datetime.time):
        if obj.utcoffset() is not None:
            raise ValueError("Cannot represent timezone-aware times.")
        r = obj.isoformat()
        return {b"__dt_time_iso__": True, b"data": r}
    elif isinstance(obj, datetime.timedelta):
        return {
            b"__dt_timedelta__": True,
            b"data": (obj.days, obj.seconds, obj.microseconds),
        }
    elif isinstance(obj, decimal.Decimal):
        return {b"__decimal__": True, b"data": str(obj)}
    elif isinstance(obj, uuid.UUID):
        return {b"__uuid__": True, b"data": str(obj)}
    else:
        return obj if chain is None else chain(obj)


def truss_msgpack_decoder(obj, chain=None):
    try:
        if b"__dt_datetime_iso__" in obj:
            return datetime.datetime.fromisoformat(obj[b"data"])
        elif b"__dt_date_iso__" in obj:
            return datetime.date.fromisoformat(obj[b"data"])
        elif b"__dt_time_iso__" in obj:
            return datetime.time.fromisoformat(obj[b"data"])
        elif b"__dt_timedelta__" in obj:
            days, seconds, microseconds = obj[b"data"]
            return datetime.timedelta(
                days=days, seconds=seconds, microseconds=microseconds
            )
        elif b"__decimal__" in obj:
            return decimal.Decimal(obj[b"data"])
        elif b"__uuid__" in obj:
            return uuid.UUID(obj[b"data"])
        else:
            return obj if chain is None else chain(obj)
    except KeyError:
        return obj if chain is None else chain(obj)


# this json object is JSONType + np.array + datetime
def is_truss_serializable(obj):
    import numpy as np

    # basic JSON types
    if isinstance(obj, (str, int, float, bool, type(None), dict, list)):
        return True
    elif isinstance(
        obj, (datetime.datetime, datetime.date, datetime.time, datetime.timedelta)
    ):
        return True
    elif isinstance(obj, np.ndarray):
        return True
    else:
        return False


def truss_msgpack_serialize(obj):
    import msgpack
    import msgpack_numpy as mp_np

    return msgpack.packb(
        obj, default=lambda x: truss_msgpack_encoder(x, chain=mp_np.encode)
    )


def truss_msgpack_deserialize(obj):
    import msgpack
    import msgpack_numpy as mp_np

    return msgpack.unpackb(
        obj, object_hook=lambda x: truss_msgpack_decoder(x, chain=mp_np.decode)
    )


class DeepNumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        import numpy as np

        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(DeepNumpyEncoder, self).default(obj)



import json  # noqa: E402
import logging  # noqa: E402
import os
import sys
from http import HTTPStatus  # noqa: E402
from threading import Thread
from typing import List

import numpy as np  # noqa: E402
import tornado.web  # noqa: E402
from common.lib_support import ensure_kfserving_installed
from common.serialization import DeepNumpyEncoder  # noqa: E402
from common.serialization import (  # noqa: E402
    truss_msgpack_deserialize,
    truss_msgpack_serialize,
)
from common.util import (  # noqa: E402
    assign_request_to_inputs_instances_after_validation,
)
from kfserving import KFModel
from kfserving.handlers.http import HTTPHandler  # noqa: E402
from kfserving.kfserver import HealthHandler, KFServer, ListHandler, LivenessHandler
from model_wrapper import ModelWrapper
from pythonjsonlogger import jsonlogger  # noqa: E402
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop

ensure_kfserving_installed()

logger = logging.getLogger(__name__)


def _configure_logging():
    json_log_handler = logging.StreamHandler()
    json_log_handler.setFormatter(
        jsonlogger.JsonFormatter("%(asctime)s %(levelname)s %(message)s")
    )
    logger = logging.getLogger()
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logger.addHandler(json_log_handler)


class TrussHTTPBinaryHandler(HTTPHandler):
    def validate(self, request):
        if (
            "instances" in request
            and not isinstance(request["instances"], (list, np.ndarray))
        ) or (
            "inputs" in request
            and not isinstance(request["inputs"], (list, np.ndarray))
        ):
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason='Expected "instances" or "inputs" to be a list or NumPy ndarray',
            )
        return assign_request_to_inputs_instances_after_validation(request)


class TrussHTTPHandler(HTTPHandler):
    def validate(self, request):
        if ("instances" in request and not isinstance(request["instances"], list)) or (
            "inputs" in request and not isinstance(request["inputs"], list)
        ):
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason='Expected "instances" or "inputs" to be a list',
            )
        return assign_request_to_inputs_instances_after_validation(request)


class TrussPredictHandler(TrussHTTPBinaryHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = truss_msgpack_deserialize(self.request.body)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e,
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.predict(request)
        response = model.postprocess(response)
        try:
            final_response = truss_msgpack_serialize(response)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e,
            )
        self.write(final_response)


class TrussExplainHandler(TrussHTTPBinaryHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = truss_msgpack_deserialize(self.request.body)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e,
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.explain(request)
        response = model.postprocess(response)
        try:
            final_response = truss_msgpack_serialize(response)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e,
            )
        self.write(final_response)


class PredictHandler(TrussHTTPHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = json.loads(self.request.body)
        except json.decoder.JSONDecodeError as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e,
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.predict(request)
        response = model.postprocess(response)
        try:
            final_response = json.dumps(response, cls=DeepNumpyEncoder)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e,
            )
        self.write(final_response)


class ExplainHandler(TrussHTTPHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = json.loads(self.request.body)
        except json.decoder.JSONDecodeError as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e,
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.explain(request)
        response = model.postprocess(response)
        try:
            final_response = json.dumps(response, cls=DeepNumpyEncoder)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e,
            )
        self.write(final_response)


class TrussServer(KFServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _configure_logging()

    def load_all(self, main_loop: IOLoop):
        try:
            for model in self.registered_models.get_models():
                model.load()
        except Exception as e:
            logging.error(f"Error loading model: {e}")

            # fixme(zack) for live reload, we don't want to kill the process?
            if not os.environ.get("CONTROL_SERVER_PORT"):
                self._http_server.stop()
                main_loop.stop()
                sys.exit(1)

    def start(self, models: List[KFModel], nest_asyncio: bool = False):
        if len(models) != 1:
            raise RuntimeError("TrussServer only supports one model")

        if not isinstance(models[0], ModelWrapper):
            raise ValueError(
                "TrussServer only accepts ModelWrapper instances instead of KFServer instances"
            )

        for model in models:
            self.register_model(model)

        self._http_server = HTTPServer(
            self.create_application(), max_buffer_size=self.max_buffer_size
        )

        logging.info("Listening on port %s", self.http_port)
        self._http_server.bind(self.http_port)

        logging.info("Will fork %d workers", self.workers)
        self._http_server.start(self.workers)

        Thread(
            target=self.load_all,
            args=[IOLoop.current()],
        ).start()

        IOLoop.current().start()

    def create_application(self):
        return tornado.web.Application(
            [
                # Server Liveness API returns 200 if server is alive.
                (r"/", LivenessHandler),
                (r"/v1/models", ListHandler, dict(models=self.registered_models)),
                # Model Health API returns 200 if model is ready to serve.
                (
                    r"/v1/models/([a-zA-Z0-9_-]+)",
                    HealthHandler,
                    dict(models=self.registered_models),
                ),
                (
                    r"/v1/models/([a-zA-Z0-9_-]+):predict",
                    PredictHandler,
                    dict(models=self.registered_models),
                ),
                (
                    r"/v1/models/([a-zA-Z0-9_-]+):predict_binary",
                    TrussPredictHandler,
                    dict(models=self.registered_models),
                ),
                (
                    r"/v1/models/([a-zA-Z0-9_-]+):explain",
                    ExplainHandler,
                    dict(models=self.registered_models),
                ),
                (
                    r"/v1/models/([a-zA-Z0-9_-]+):explain_binary",
                    TrussExplainHandler,
                    dict(models=self.registered_models),
                ),
            ]
        )



import os
import random
import string
import subprocess as sp
import tempfile
from contextlib import contextmanager
from distutils.dir_util import copy_tree, remove_tree
from distutils.file_util import copy_file
from pathlib import Path


def copy_tree_path(src: Path, dest: Path):
    return copy_tree(str(src), str(dest))


def copy_file_path(src: Path, dest: Path):
    return copy_file(str(src), str(dest))


def remove_tree_path(target: Path):
    return remove_tree(str(target))


def get_max_modified_time_of_dir(path: Path) -> float:
    max_modified_time = os.path.getmtime(path)
    for root, dirs, files in os.walk(path):
        if os.path.islink(root):
            raise ValueError(f"Symlinks not allowed in Truss: {root}")
        files = [f for f in files if not f.startswith(".")]
        dirs[:] = [d for d in dirs if not d.startswith(".")]
        max_modified_time = max(max_modified_time, os.path.getmtime(root))
        for file in files:
            max_modified_time = max(
                max_modified_time, os.path.getmtime(os.path.join(root, file))
            )
    return max_modified_time


@contextmanager
def given_or_temporary_dir(given_dir: Path = None):
    if given_dir is not None:
        yield given_dir
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)


def build_truss_target_directory(stub: str) -> Path:
    """Builds a directory under ~/.truss/models for the purpose of creating a Truss at."""
    rand_suffix = "".join(random.choices(string.ascii_uppercase + string.digits, k=6))
    target_directory_path = Path(
        Path.home(), ".truss", "models", f"{stub}-{rand_suffix}"
    )
    target_directory_path.mkdir(parents=True)
    return target_directory_path


def get_gpu_memory():
    # https://stackoverflow.com/questions/59567226/how-to-programmatically-determine-available-gpu-memory-with-tensorflow
    try:
        command = "nvidia-smi --query-gpu=memory.used --format=csv"
        memory_free_info = (
            sp.check_output(command.split()).decode("ascii").split("\n")[1]
        )
        memory_free_values = int(memory_free_info.split()[0])
        return memory_free_values
    except FileNotFoundError:
        return None
